// Microsoft Visual C++ generated resource script.
//
#include "resource.h"

#define APSTUDIO_READONLY_SYMBOLS
/////////////////////////////////////////////////////////////////////////////
//
// Generated from the TEXTINCLUDE 2 resource.
//
#ifndef APSTUDIO_INVOKED
#include "targetver.h"
#endif
#define APSTUDIO_HIDDEN_SYMBOLS
#include "windows.h"
#undef APSTUDIO_HIDDEN_SYMBOLS

/////////////////////////////////////////////////////////////////////////////
#undef APSTUDIO_READONLY_SYMBOLS

/////////////////////////////////////////////////////////////////////////////
// English (United States) resources

#if !defined(AFX_RESOURCE_DLL) || defined(AFX_TARG_ENU)
LANGUAGE LANG_ENGLISH, SUBLANG_ENGLISH_US

/////////////////////////////////////////////////////////////////////////////
//
// Icon
//

// Icon with lowest ID value placed first to ensure application icon
// remains consistent on all systems.
IDI_STITCH              ICON                    "Stitch.ico"
IDI_SMALL               ICON                    "small.ico"

/////////////////////////////////////////////////////////////////////////////
//
// Menu
//

IDC_STITCH MENU
BEGIN
    POPUP "&File"
    BEGIN
        MENUITEM "Add File... \tO",             ID_FILE_OPEN
        MENUITEM "Clear File List",             ID_FILE_CLEARFILELIST
        MENUITEM "Open Project...",             ID_FILE_OPENPTGUIPROJECT
        MENUITEM SEPARATOR
        MENUITEM "ZCam Live (1080P)",           ID_FILE_ZCAMLIVE
        MENUITEM "ZCam Live (2.7K)",            ID_FILE_ZCAMLIVE_2K7
        MENUITEM "ZCam Live (2544P)",           ID_FILE_ZCAMLIVE_2544P
        MENUITEM SEPARATOR
        MENUITEM "Ricoh Theta S Live",          ID_FILE_RICOH_THETA_S
        MENUITEM "Ricoh Theta V Live",          ID_FILE_RICOH_THETA_V
        MENUITEM SEPARATOR
        MENUITEM "E&xit",                       IDM_EXIT
    END
    POPUP "&Options"
    BEGIN
        MENUITEM "Renderer DirectCompute",      ID_OPTIONS_RENDERER_DX11
        MENUITEM "Renderer Compute For DX11",   ID_OPTIONS_COMPUTE_FOR_DX11
        MENUITEM "Renderer OpenCL",             ID_OPTIONS_RENDERER_OPENCL
        MENUITEM SEPARATOR
        MENUITEM "Loop",                        ID_OPTIONS_LOOP, CHECKED
        MENUITEM "Force 60 FPS",                ID_OPTIONS_FORCE60FPS, CHECKED
        MENUITEM "Full Speed",                  ID_OPTIONS_FULLSPEED
        MENUITEM "Low Latency",                 ID_OPTIONS_LOW_LATENCY
        MENUITEM SEPARATOR
    END
    MENUITEM "Play",                        ID_PLAYBACK_PLAY
    MENUITEM "Pause",                       ID_PLAYBACK_PAUSE
    MENUITEM "Step",                        ID_PLAYBACK_STEP
    MENUITEM "Stop",                        ID_PLAYBACK_STOP
    POPUP "&Help"
    BEGIN
        MENUITEM "&About ...",                  IDM_ABOUT
    END
END


/////////////////////////////////////////////////////////////////////////////
//
// Accelerator
//

IDC_STITCH ACCELERATORS
BEGIN
    "?",            IDM_ABOUT,              ASCII,  ALT
    "/",            IDM_ABOUT,              ASCII,  ALT
END


/////////////////////////////////////////////////////////////////////////////
//
// Dialog
//

IDD_PROGRESS_DLG DIALOGEX 0, 0, 199, 86
STYLE DS_SETFONT | DS_MODALFRAME | DS_FIXEDSYS | WS_POPUP | WS_CAPTION | WS_SYSMENU
CAPTION "Playback"
FONT 8, "MS Shell Dlg", 0, 0, 0x1
BEGIN
    CONTROL         "",IDC_PROGRESS_BAR,"msctls_progress32",WS_BORDER,12,38,173,14
    LTEXT           "Optimization Components...",IDC_STATIC,12,21,90,8
    GROUPBOX        "Optimization Media SDK for your hardware",IDC_STATIC,4,4,189,57
    DEFPUSHBUTTON   "&Cancel",IDCANCEL,143,66,50,14
END

IDD_ABOUTBOX DIALOGEX 0, 0, 170, 62
STYLE DS_SETFONT | DS_MODALFRAME | DS_FIXEDSYS | WS_POPUP | WS_CAPTION | WS_SYSMENU
CAPTION "About Stitch"
FONT 8, "MS Shell Dlg", 0, 0, 0x1
BEGIN
    ICON            IDR_MAINFRAME,IDC_STATIC,14,14,21,20
    LTEXT           "Stitch, Version 1.0",IDC_STATIC,42,14,114,8,SS_NOPREFIX
    LTEXT           "Copyright (C) 2015",IDC_STATIC,42,26,114,8
    DEFPUSHBUTTON   "OK",IDOK,113,41,50,14,WS_GROUP
END


/////////////////////////////////////////////////////////////////////////////
//
// DESIGNINFO
//

#ifdef APSTUDIO_INVOKED
GUIDELINES DESIGNINFO
BEGIN
    IDD_ABOUTBOX, DIALOG
    BEGIN
        LEFTMARGIN, 7
        RIGHTMARGIN, 163
        TOPMARGIN, 7
        BOTTOMMARGIN, 55
    END
END
#endif    // APSTUDIO_INVOKED


#ifdef APSTUDIO_INVOKED
/////////////////////////////////////////////////////////////////////////////
//
// TEXTINCLUDE
//

1 TEXTINCLUDE 
BEGIN
    "resource.h\0"
END

2 TEXTINCLUDE 
BEGIN
    "#ifndef APSTUDIO_INVOKED\r\n"
    "#include ""targetver.h""\r\n"
    "#endif\r\n"
    "#define APSTUDIO_HIDDEN_SYMBOLS\r\n"
    "#include ""windows.h""\r\n"
    "#undef APSTUDIO_HIDDEN_SYMBOLS\r\n"
    "\0"
END

3 TEXTINCLUDE 
BEGIN
    "\r\n"
    "\0"
END

#endif    // APSTUDIO_INVOKED


/////////////////////////////////////////////////////////////////////////////
//
// String Table
//

STRINGTABLE
BEGIN
    IDS_APP_TITLE           "Stitch"
    IDC_STITCH              "STITCH"
END

#endif    // English (United States) resources
/////////////////////////////////////////////////////////////////////////////



#ifndef APSTUDIO_INVOKED
/////////////////////////////////////////////////////////////////////////////
//
// Generated from the TEXTINCLUDE 3 resource.
//


/////////////////////////////////////////////////////////////////////////////
#endif    // not APSTUDIO_INVOKED

