// Microsoft Visual C++ generated resource script.
//
#include "resource.h"

#define APSTUDIO_READONLY_SYMBOLS
/////////////////////////////////////////////////////////////////////////////
//
// Generated from the TEXTINCLUDE 2 resource.
//
#ifndef APSTUDIO_INVOKED
#include "targetver.h"
#endif
#define APSTUDIO_HIDDEN_SYMBOLS
#include "windows.h"
#undef APSTUDIO_HIDDEN_SYMBOLS

/////////////////////////////////////////////////////////////////////////////
#undef APSTUDIO_READONLY_SYMBOLS

/////////////////////////////////////////////////////////////////////////////
// English (United States) resources

#if !defined(AFX_RESOURCE_DLL) || defined(AFX_TARG_ENU)
LANGUAGE LANG_ENGLISH, SUBLANG_ENGLISH_US

/////////////////////////////////////////////////////////////////////////////
//
// Icon
//

// Icon with lowest ID value placed first to ensure application icon
// remains consistent on all systems.
IDI_CAPTUREVIDEO        ICON                    "CaptureVideo.ico"

IDI_SMALL               ICON                    "small.ico"


/////////////////////////////////////////////////////////////////////////////
//
// Menu
//

IDC_CAPTUREVIDEO MENU
BEGIN
    POPUP "&File"
    BEGIN
        MENUITEM "Media File... \tF",           ID_FILE_OPEN
        POPUP "Capture Device"
        BEGIN
            MENUITEM "RICOH THETA S",               ID_VIDEO_CAPTURE0
            MENUITEM "RICOH THETA V",               ID_VIDEO_CAPTURE1
            MENUITEM "Other2",                      ID_VIDEO_CAPTURE2
            MENUITEM "Other3",                      ID_VIDEO_CAPTURE3
            MENUITEM "Other4",                      ID_VIDEO_CAPTURE4
            MENUITEM "Other5",                      ID_VIDEO_CAPTURE5
            MENUITEM "Other6",                      ID_VIDEO_CAPTURE6
            MENUITEM "Other7",                      ID_VIDEO_CAPTURE7
            MENUITEM "Other8",                      ID_VIDEO_CAPTURE8
            MENUITEM "Other9",                      ID_VIDEO_CAPTURE9
        END
        MENUITEM SEPARATOR
        MENUITEM "BackGround Media File... \tB", ID_FILE_OPEN_BK
        MENUITEM SEPARATOR
        MENUITEM "Load Options",                ID_FILE_LOAD_OPTIONS
        MENUITEM "Save Options",                ID_FILE_SAVE_OPTIONS
        MENUITEM "Reset Options",               ID_FILE_RESETHISTORY
        MENUITEM "E&xit",                       ID_FILE_EXIT
    END
    POPUP "&Options"
    BEGIN
        MENUITEM SEPARATOR
        MENUITEM "Loop",                        ID_OPTIONS_LOOP, CHECKED
        MENUITEM SEPARATOR
        MENUITEM "Chroma Key",                  ID_OPTIONS_CHROMAKEY
        MENUITEM "Blend Background",            ID_OPTIONS_CHROMAKEY_BK
        MENUITEM "Zoom",                        ID_OPTIONS_SCALING
        MENUITEM SEPARATOR
        MENUITEM "Spill suppression",           ID_OPTIONS_CHROMAKEY_SPILL
        MENUITEM "Color Adjustment",            ID_OPTIONS_CHROMAKEY_COLOR
        MENUITEM "Advanced",                    ID_OPTIONS_CHROMAKEY_ADVANCED
        MENUITEM SEPARATOR
        MENUITEM "10 bit Capture",              ID_OPTIONS_CHROMAKEY_10BIT
        MENUITEM "RGBA_FP16 output",            ID_OPTIONS_CHROMAKEY_RGBAFP16
        MENUITEM "alpha from source",           ID_OPTIONS_CHROMAKEY_ALPHA_SRC
        MENUITEM "Debug",                       ID_OPTIONS_DEBUG
    END
    MENUITEM "Run",                         ID_PLAYBACK_PLAY
    MENUITEM "Pause",                       ID_PLAYBACK_PAUSE
    MENUITEM "Step",                        ID_PLAYBACK_STEP
    MENUITEM "Stop",                        ID_PLAYBACK_STOP
    POPUP "&Help"
    BEGIN
        MENUITEM "&About ...",                  IDM_ABOUT
    END
END


/////////////////////////////////////////////////////////////////////////////
//
// Accelerator
//

IDC_CAPTUREVIDEO ACCELERATORS
BEGIN
    "?",            IDM_ABOUT,              ASCII,  ALT
    "/",            IDM_ABOUT,              ASCII,  ALT
END


/////////////////////////////////////////////////////////////////////////////
//
// Dialog
//

IDD_PROGRESS_DLG DIALOGEX 0, 0, 199, 86
STYLE DS_SETFONT | DS_MODALFRAME | DS_FIXEDSYS | WS_POPUP | WS_CAPTION | WS_SYSMENU
CAPTION "Playback"
FONT 8, "MS Shell Dlg", 0, 0, 0x1
BEGIN
    DEFPUSHBUTTON   "&Cancel",IDCANCEL,143,66,50,14
    CONTROL         "",IDC_PROGRESS_BAR,"msctls_progress32",WS_BORDER,12,38,173,14
    LTEXT           "Optimization Components...",IDC_STATIC,12,21,90,8
    GROUPBOX        "Optimization Media SDK for your hardware",IDC_STATIC,4,4,189,57
END

IDD_ABOUTBOX DIALOGEX 0, 0, 170, 62
STYLE DS_SETFONT | DS_MODALFRAME | DS_FIXEDSYS | WS_POPUP | WS_CAPTION | WS_SYSMENU
CAPTION "About CaptureVideo"
FONT 8, "MS Shell Dlg", 0, 0, 0x1
BEGIN
    DEFPUSHBUTTON   "OK",IDOK,113,41,50,14,WS_GROUP
    LTEXT           "Copyright (C) 2019",IDC_STATIC,42,26,114,8
    ICON            128,IDC_STATIC,14,14,20,20
    LTEXT           "Capture Video, Version 1.0",IDC_STATIC,42,14,114,8,SS_NOPREFIX
END

IDD_PROPERTY_DIALOG DIALOGEX 0, 0, 316, 182
STYLE DS_SETFONT | DS_MODALFRAME | DS_FIXEDSYS | WS_POPUP | WS_VISIBLE | WS_CAPTION | WS_SYSMENU
CAPTION "Properties"
FONT 8, "MS Shell Dlg", 400, 0, 0x1
BEGIN
    CONTROL         "a",IDC_CURVE_EDITOR,"Button",BS_OWNERDRAW | WS_TABSTOP,7,7,242,168
    PUSHBUTTON      "Add Point",IDC_BUTTON_ADD_POINT,259,7,50,14
    PUSHBUTTON      "DeletePoint",IDC_BUTTON_DELETE_POINT,259,22,50,14
    PUSHBUTTON      "Reset",IDC_BUTTON_RESET,259,42,50,14
    DEFPUSHBUTTON   "Close",IDOK,259,161,50,14
END


/////////////////////////////////////////////////////////////////////////////
//
// DESIGNINFO
//

#ifdef APSTUDIO_INVOKED
GUIDELINES DESIGNINFO
BEGIN
    IDD_PROGRESS_DLG, DIALOG
    BEGIN
    END

    IDD_ABOUTBOX, DIALOG
    BEGIN
        LEFTMARGIN, 7
        RIGHTMARGIN, 163
        TOPMARGIN, 7
        BOTTOMMARGIN, 55
    END

    IDD_PROPERTY_DIALOG, DIALOG
    BEGIN
        LEFTMARGIN, 7
        RIGHTMARGIN, 309
        TOPMARGIN, 7
        BOTTOMMARGIN, 175
    END
END
#endif    // APSTUDIO_INVOKED


#ifdef APSTUDIO_INVOKED
/////////////////////////////////////////////////////////////////////////////
//
// TEXTINCLUDE
//

1 TEXTINCLUDE 
BEGIN
    "resource.h\0"
END

2 TEXTINCLUDE 
BEGIN
    "#ifndef APSTUDIO_INVOKED\r\n"
    "#include ""targetver.h""\r\n"
    "#endif\r\n"
    "#define APSTUDIO_HIDDEN_SYMBOLS\r\n"
    "#include ""windows.h""\r\n"
    "#undef APSTUDIO_HIDDEN_SYMBOLS\r\n"
    "\0"
END

3 TEXTINCLUDE 
BEGIN
    "\r\n"
    "\0"
END

#endif    // APSTUDIO_INVOKED


/////////////////////////////////////////////////////////////////////////////
//
// String Table
//

STRINGTABLE
BEGIN
    IDS_APP_TITLE           "CaptureVideo"
    IDC_CAPTUREVIDEO        "CaptureVideo"
END

#endif    // English (United States) resources
/////////////////////////////////////////////////////////////////////////////



#ifndef APSTUDIO_INVOKED
/////////////////////////////////////////////////////////////////////////////
//
// Generated from the TEXTINCLUDE 3 resource.
//


/////////////////////////////////////////////////////////////////////////////
#endif    // not APSTUDIO_INVOKED

