//{{NO_DEPENDENCIES}}
// Microsoft Visual C++ generated include file.
// Used by CaptureVideo.rc
//
#define IDS_APP_TITLE                   103
#define IDD_ABOUTBOX                    103
#define IDM_ABOUT                       104
#define IDI_CAPTUREVIDEO                107
#define IDI_SMALL                       108
#define IDC_CAPTUREVIDEO                109
#define IDD_PROGRESS_DLG                131
#define IDD_PROPERTY_DIALOG             132
#define IDC_PROGRESS_BAR                1000
#define IDC_CURVE_EDITOR                1001
#define IDC_BUTTON_ADD_POINT            1002
#define IDC_BUTTON_DELETE_POINT         1003
#define IDC_BUTTON_RESET                1004
#define ID_FILE_OPEN                    32771
#define ID_PLAYBACK_PLAY                32785
#define ID_PLAYBACK_PAUSE               32786
#define ID_PLAYBACK_STEP                32788
#define ID_PLAYBACK_STOP                32789
#define ID_OPTIONS_LOOP                 32790
#define ID_FILE_EXIT                    32832
#define ID_FILE_RESETHISTORY            32835
#define ID_VIDEO_CAPTURE0               32852
#define ID_VIDEO_CAPTURE1               32853
#define ID_VIDEO_CAPTURE2               32854
#define ID_VIDEO_CAPTURE3               32855
#define ID_VIDEO_CAPTURE4               32856
#define ID_VIDEO_CAPTURE5               32857
#define ID_VIDEO_CAPTURE6               32858
#define ID_VIDEO_CAPTURE7               32859
#define ID_VIDEO_CAPTURE8               32860
#define ID_VIDEO_CAPTURE9               32861
#define ID_FILE_OPEN_BK                 32864
#define ID_OPTIONS_CHROMAKEY            32865
#define ID_OPTIONS_CHROMAKEY_BK         32866
#define ID_OPTIONS_DEBUG                32867
#define ID_OPTIONS_CHROMAKEY_COLOR      32868
#define ID_OPTIONS_CHROMAKEY_SPILL      32869
#define ID_OPTIONS_CHROMAKEY_ADVANCED   32870
#define ID_OPTIONS_SCALING              32871
#define ID_OPTIONS_CHROMAKEY_RGBAFP16   32872
#define ID_OPTIONS_CHROMAKEY_10BIT      32873
#define ID_FILE_LOAD_OPTIONS            32874
#define ID_FILE_SAVE_OPTIONS            32875
#define ID_OPTIONS_CHROMAKEY_ALPHA_SRC  32876
#define IDC_STATIC                      -1

// Next default values for new objects
// 
#ifdef APSTUDIO_INVOKED
#ifndef APSTUDIO_READONLY_SYMBOLS
#define _APS_NO_MFC                     1
#define _APS_NEXT_RESOURCE_VALUE        134
#define _APS_NEXT_COMMAND_VALUE         32877
#define _APS_NEXT_CONTROL_VALUE         1062
#define _APS_NEXT_SYMED_VALUE           110
#endif
#endif
