//{{NO_DEPENDENCIES}}
// Microsoft Visual C++ generated include file.
// Used by PlaybackHW.rc
//
#define IDC_MYICON                      2
#define IDD_PLAYBACKHW_DIALOG           102
#define IDS_APP_TITLE                   103
#define IDD_ABOUTBOX                    103
#define IDM_ABOUT                       104
#define IDM_EXIT                        105
#define IDI_PLAYBACKHW                  107
#define IDI_SMALL                       108
#define IDC_PLAYBACKHW                  109
#define IDR_MAINFRAME                   128
#define IDD_TOOLBAR_DLG                 130
#define IDD_PROGRESS_DLG                131
#define IDD_OPEN_STREAM                 136
#define IDC_BRIGHTNESS_EDIT             1001
#define IDC_BRIGHTNESS_SPIN             1002
#define IDC_CONTRAST_EDIT               1003
#define IDC_CONTRAST_SPIN               1004
#define IDC_HUE_EDIT                    1005
#define IDC_HUE_SPIN                    1006
#define IDC_SATURATION_EDIT             1007
#define IDC_SATURATION_SPIN             1008
#define IDC_GAMMA_EDIT                  1009
#define IDC_GAMMA_SPIN                  1010
#define IDC_DYNAMIC_RANGE_COMBO         1011
#define IDC_DEINTERLACING_MODE_COMBO    1012
#define IDC_PULDOWN_DETECTION_CHECK     1013
#define IDC_TEMPORAL_DENOISE_CHECK      1014
#define IDC_TEMPORAL_DENOISE_EDIT       1015
#define IDC_PROGRESS_BAR                1015
#define IDC_TEMPORAL_DENOISE_SPIN       1016
#define IDC_SEEK                        1016
#define IDC_DEBLOCKING_CHECK            1017
#define IDC_CHECK1                      1017
#define IDC_CHECK_LISTEN                1018
#define IDC_DEBLOCKING_EDIT             1018
#define IDC_DEBLOCKING_SPIN             1019
#define IDC_MOSQUITO_NOISE_REDUCTION_CHECK 1020
#define IDC_MOSQUITO_NOISE_REDUCTION_EDIT 1021
#define IDC_MOSQUITO_NOISE_REDUCTION_SPIN 1022
#define IDC_FALSE_CONTOURS_REMOVAL_CHECK 1023
#define IDC_FALSE_CONTOURS_REMOVAL_EDIT 1024
#define IDC_FALSE_CONTOURS_REMOVAL_SPIN 1025
#define IDC_COLOR_VIBRANCE_CHECK        1026
#define IDC_COLOR_VIBRANCE_EDIT         1027
#define IDC_COLOR_VIBRANCE_SPIN         1028
#define IDC_EDIT_URL                    1028
#define IDC_EDIT_URL_VIDEO              1028
#define IDC_SKINTONE_CORRECTION_CHECK   1029
#define IDC_EDIT_URL2                   1029
#define IDC_EDIT_URL_AUDIO              1029
#define IDC_SKINTONE_CORRECTION_EDIT    1030
#define IDC_SKINTONE_CORRECTION_SPIN    1031
#define IDC_BRIGHTER_WHITES_CHECK       1032
#define IDC_DEINTERLACING_MODE_COMBO2   1033
#define IDC_PULLDOWN_MODE_COMBO         1033
#define IDC_DEBLOCKING_CHECK2           1033
#define IDC_STEADY_VIDEO_CHECK          1033
#define IDC_EDGE_ENHANCEMENT_CHECK      1035
#define IDC_EDGE_ENHANCEMENT_EDIT       1036
#define IDC_EDGE_ENHANCEMENT_SPIN       1037
#define IDC_DYNAMIC_CONTRAST_CHECK      1038
#define IDC_DEMO_MODE_CHECK             1039
#define IDC_FRAME_RATE_CONVERSION_CHECK 1040
#define IDC_STEADY_VIDEO_STRENGTH_EDIT  1044
#define IDC_STEADY_VIDEO_STRENGTH_SPIN  1045
#define IDC_STEADY_VIDEO_DELAY_EDIT     1046
#define IDC_STEADY_VIDEO_DELAY_SPIN     1047
#define IDC_STEADY_VIDEO_ZOOM_EDIT      1048
#define IDC_STEADY_VIDEO_ZOOM_SPIN      1049
#define ID_FILE_OPEN                    32771
#define ID_Menu                         32772
#define ID_OPTIONS_PRESENTER_DX11       32780
#define ID_OPTIONS_PRESENTER_DX9        32781
#define ID_OPTIONS_PRESENTER_OPENGL     32782
#define ID_OPTIONS_PRESENTER_DX12       32783
#define ID_OPTIONS_PRESENTER_VULKAN     32784
#define ID_PLAYBACK_PLAY                32785
#define ID_PLAYBACK_PAUSE               32786
#define ID_PLAYBACK_STEP                32788
#define ID_PLAYBACK_STOP                32789
#define ID_TOOLBAR                      32793
#define ID_FILE_OPENNETWORKSTREAM       32794
#define ID_OPTIONS_FULLSCREEN           32797
#define ID_OPTIONS_HQ_SCALER            32798
#define ID_HQSCALER_OFF                 32799
#define ID_HQSCALER_BILINEAR            32800
#define ID_HQSCALER_BICUBIC             32801
#define ID_HQSCALER_FSR                 32802
#define ID_OPTIONS_EXCLUSIVEFULLSCREEN  32807
#define ID_OPTIONS_FRC                  32808
#define IDC_STATIC                      -1

// Next default values for new objects
// 
#ifdef APSTUDIO_INVOKED
#ifndef APSTUDIO_READONLY_SYMBOLS
#define _APS_NO_MFC                     1
#define _APS_NEXT_RESOURCE_VALUE        131
#define _APS_NEXT_COMMAND_VALUE         32809
#define _APS_NEXT_CONTROL_VALUE         1018
#define _APS_NEXT_SYMED_VALUE           110
#endif
#endif
