// 
// Notice Regarding Standards.  AMD does not provide a license or sublicense to
// any Intellectual Property Rights relating to any standards, including but not
// limited to any audio and/or video codec technologies such as MPEG-2, MPEG-4;
// AVC/H.264; HEVC/H.265; AAC decode/FFMPEG; AAC encode/FFMPEG; VC-1; and MP3
// (collectively, the "Media Technologies"). For clarity, you will pay any
// royalties due for such third party technologies, which may include the Media
// Technologies that are owed as a result of AMD providing the Software to you.
// 
// MIT license 
// 
// Copyright (c) 2017 Advanced Micro Devices, Inc. All rights reserved.
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.
//

#pragma once

#include <atlcomcli.h>

#include <windows.h>
#include <windowsx.h>
#include <mfapi.h>
#include <mfidl.h>
#include <mferror.h>
#include <mfcaptureengine.h>
#include <shlwapi.h>
#include <strsafe.h>
#include <commctrl.h>
#include <vector>
#include <string>
#include <mmdeviceapi.h>
#include <Mfreadwrite.h>

#include "public/include/core/Data.h"

namespace amf
{
    class AMFMFSourceImpl
    {
    public:
        AMFMFSourceImpl();
        virtual ~AMFMFSourceImpl();

        AMF_RESULT Close();
        AMF_RESULT  Init(std::wstring& deviceNameActive);
        amf_uint32  GetFrameWidth(){ return m_width; };
        amf_uint32  GetFrameHeight(){ return m_height; };
        GUID        GetCodec(){ return m_guidSubType; };
        amf_pts     GetFrameDuration(){ return m_duration; };
        amf_uint32  GetDeviceCount() { return (amf_uint32)m_deviceList.size(); };
        std::string GetDeviceName(amf_uint32 index)
                   {return (index >= m_deviceList.size()) ? std::string("") : m_deviceList[index];};
        AMF_RESULT  CaptureOnePacket(char** ppData, UINT& numSamples);
        AMF_RESULT  CaptureOnePacketDone();

    protected:
        AMF_RESULT CreateDeviceList();
        AMF_RESULT    SelectSource(CComPtr<IMFMediaSource> pMediaSource);

        CComPtr<IMFMediaSource>  m_pMediaSource;
        CComPtr<IMFSourceReader> m_pReader;
        CComPtr<IMFSample>       m_pSample;
        CComPtr<IMFMediaBuffer>  m_pMediaBuf;

        std::vector<std::string>  m_deviceList;
        MFVIDEOFORMAT             m_videoFormat;
        GUID                      m_guidSubType;
        amf_uint32                m_width;
        amf_uint32                m_height;
        amf_pts                   m_duration;
        amf_uint32                m_selected;
    };
    typedef AMFInterfacePtr_T<AMFMFSourceImpl>    AMFMFSourceImplPtr;
} //namespace amf
