// 
// Notice Regarding Standards.  AMD does not provide a license or sublicense to
// any Intellectual Property Rights relating to any standards, including but not
// limited to any audio and/or video codec technologies such as MPEG-2, MPEG-4;
// AVC/H.264; HEVC/H.265; AAC decode/FFMPEG; AAC encode/FFMPEG; VC-1; and MP3
// (collectively, the "Media Technologies"). For clarity, you will pay any
// royalties due for such third party technologies, which may include the Media
// Technologies that are owed as a result of AMD providing the Software to you.
// 
// MIT license 
// 
//
// Copyright (c) 2018 Advanced Micro Devices, Inc. All rights reserved.
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.
//

#include "VideoRender.h"
#if defined(_WIN32)
#include "VideoRenderOpenGL.h"
#include "VideoRenderOpenCL.h"
#include "VideoRenderDX9.h"
#include "VideoRenderDX11.h"
#include "VideoRenderDX12.h"
#endif
#include "VideoRenderHost.h"
#include "VideoRenderVulkan.h"

VideoRender::VideoRender(amf_int width, amf_int height, bool bInterlaced, amf_int frames, amf::AMFContext* pContext)
    :m_pContext(pContext),
    m_width(width),
    m_height(height),
    m_frames(frames),
    m_framesRendered(0),
    m_bInterlaced(bInterlaced),
	m_renderFps(0),
	m_lastRenderTime(0)
{

}

VideoRender::~VideoRender()
{
}

VideoRenderPtr VideoRender::Create(amf_int width, amf_int height, bool bInterlaced, amf_int frames, amf::AMF_MEMORY_TYPE type, amf::AMF_MEMORY_TYPE encodertype, amf::AMFContext* pContext)
{
    VideoRenderPtr engine;
    switch(type)
    {
#if defined(_WIN32)
    case amf::AMF_MEMORY_DX9:
        engine = VideoRenderPtr(new VideoRenderDX9(width, height, bInterlaced, frames, pContext));
        break;
    case amf::AMF_MEMORY_DX11:
        engine = VideoRenderPtr(new VideoRenderDX11(width, height, bInterlaced, frames, pContext));
        break;
    case amf::AMF_MEMORY_DX12:
        engine = VideoRenderPtr(new VideoRenderDX12(width, height, bInterlaced, frames, pContext));
        break;
    case amf::AMF_MEMORY_OPENGL:
        engine = VideoRenderPtr(new VideoRenderOpenGL(width, height, bInterlaced, frames, pContext));
        break;
    case amf::AMF_MEMORY_OPENCL:
        engine = VideoRenderPtr(new VideoRenderOpenCL(width, height, bInterlaced, frames, pContext, encodertype));
        break;
#endif
    case amf::AMF_MEMORY_HOST:
        engine = VideoRenderPtr(new VideoRenderHost(width, height, bInterlaced, frames, pContext));
        break;
    case amf::AMF_MEMORY_VULKAN:
        engine = VideoRenderPtr(new VideoRenderVulkan(width, height, bInterlaced, frames, pContext));
        break;
    default:
        break;
    }
    return engine;
}


AMF_RESULT VideoRender::QueryOutput(amf::AMFData** ppData)
{
    if(m_framesRendered++ < m_frames)
    {
		if (m_renderFps > 0) //render with the fps setting.
		{
			amf_int64 currentTime = amf_high_precision_clock();
			amf::AMFPreciseWaiter waiter;
			amf_int64 waittime = (AMF_SECOND / m_renderFps) - (currentTime - m_lastRenderTime);
			if(waittime>0)
			{
				waiter.WaitEx(waittime);
			}
			m_lastRenderTime = amf_high_precision_clock();
		}

        AMF_RESULT res = Render(ppData);
        if(res == AMF_OK && *ppData != NULL && m_bInterlaced)
        {
            amf::AMFSurfacePtr(*ppData)->SetFrameType(amf::AMF_FRAME_INTERLEAVED_EVEN_FIRST);
        }
        return res;
    }
    return AMF_EOF;
}
