// Microsoft Visual C++ generated resource script.
//
#include "resource.h"
#define APSTUDIO_READONLY_SYMBOLS
/////////////////////////////////////////////////////////////////////////////
//
// Generated from the TEXTINCLUDE 2 resource.
//
#include "winres.h"

/////////////////////////////////////////////////////////////////////////////
#undef APSTUDIO_READONLY_SYMBOLS

/////////////////////////////////////////////////////////////////////////////
// English (United States) resources

#if !defined(AFX_RESOURCE_DLL) || defined(AFX_TARG_ENU)
LANGUAGE LANG_ENGLISH, SUBLANG_ENGLISH_US
#pragma code_page(1252)

#ifdef APSTUDIO_INVOKED
/////////////////////////////////////////////////////////////////////////////
//
// TEXTINCLUDE
//

1 TEXTINCLUDE
BEGIN
    "resource.h\0"
END

2 TEXTINCLUDE
BEGIN
    "#include ""winres.h""\r\n"
    "\0"
END

3 TEXTINCLUDE
BEGIN
    "\r\n"
    "\0"
END

#endif    // APSTUDIO_INVOKED


/////////////////////////////////////////////////////////////////////////////
//
// Icon
//

// Icon with lowest ID value placed first to ensure application icon
// remains consistent on all systems.
IDI_DVR                 ICON                    "DVR.ico"


/////////////////////////////////////////////////////////////////////////////
//
// Menu
//

IDC_DVR MENU
BEGIN
    POPUP "&File"
    BEGIN
        MENUITEM "Change Save Location",        IDM_SAVEFILE
        MENUITEM SEPARATOR
        MENUITEM "E&xit",                       IDM_EXIT
    END
    POPUP "&Devices"
    BEGIN
        MENUITEM SEPARATOR
    END
    POPUP "Capture &Source"
    BEGIN
        MENUITEM SEPARATOR
        MENUITEM "Multi-monitor",               ID_CAPTURESOURCE_MULTI_MONITOR
    END
    POPUP "Capture Component"
    BEGIN
        MENUITEM SEPARATOR
    END
    POPUP "Processing"
    BEGIN
        MENUITEM SEPARATOR
        MENUITEM "Pre-Analysis",               ID_PRE_ANALYSIS
    END
    POPUP "&Help"
    BEGIN
        MENUITEM "&About ...",                  IDM_ABOUT
    END
END


/////////////////////////////////////////////////////////////////////////////
//
// Dialog
//

IDD_ABOUTBOX DIALOGEX 0, 0, 185, 61
STYLE DS_SETFONT | DS_MODALFRAME | DS_FIXEDSYS | WS_POPUP | WS_CAPTION | WS_SYSMENU
CAPTION "About DVR"
FONT 8, "MS Shell Dlg", 400, 0, 0x1
BEGIN
    DEFPUSHBUTTON   "OK",IDOK,129,41,50,14
    ICON            IDI_DVR,IDC_STATIC,18,20,20,20
    LTEXT           "AMF DVR Sample, Version 1.0",IDC_STATIC,46,19,99,10
    LTEXT           "Copyright (C) 2017",IDC_STATIC,46,30,65,8
END

IDD_NODEVICEBOX DIALOGEX 0, 0, 223, 73
STYLE DS_SETFONT | DS_MODALFRAME | DS_FIXEDSYS | WS_POPUP | WS_CAPTION | WS_SYSMENU
CAPTION "No Devices Detected"
FONT 8, "MS Shell Dlg", 400, 0, 0x1
BEGIN
    DEFPUSHBUTTON   "Close",IDCLOSE,166,52,50,14
    CTEXT           "No valid AMD GPU devices have been detected. The DVR API requires an AMD device.",IDC_STATIC,30,23,163,20
END


/////////////////////////////////////////////////////////////////////////////
//
// DESIGNINFO
//

#ifdef APSTUDIO_INVOKED
GUIDELINES DESIGNINFO
BEGIN
    IDD_ABOUTBOX, DIALOG
    BEGIN
        LEFTMARGIN, 6
        RIGHTMARGIN, 179
        TOPMARGIN, 6
        BOTTOMMARGIN, 55
    END

    IDD_NODEVICEBOX, DIALOG
    BEGIN
        LEFTMARGIN, 7
        RIGHTMARGIN, 216
        TOPMARGIN, 7
        BOTTOMMARGIN, 66
    END
END
#endif    // APSTUDIO_INVOKED


/////////////////////////////////////////////////////////////////////////////
//
// String Table
//

STRINGTABLE
BEGIN
    IDC_DVR                 "AMF DVR Sample"
    IDS_APP_TITLE           "AMF DVR Sample"
END

#endif    // English (United States) resources
/////////////////////////////////////////////////////////////////////////////


/////////////////////////////////////////////////////////////////////////////
// English (Canada) resources

#if !defined(AFX_RESOURCE_DLL) || defined(AFX_TARG_ENC)
LANGUAGE LANG_ENGLISH, SUBLANG_ENGLISH_CAN
#pragma code_page(1252)

/////////////////////////////////////////////////////////////////////////////
//
// Dialog
//

IDD_MAIN_DIALOG DIALOGEX 0, 0, 255, 150
STYLE DS_SETFONT | DS_MODALFRAME | DS_FIXEDSYS | WS_POPUP | WS_VISIBLE | WS_CAPTION | WS_SYSMENU
EXSTYLE WS_EX_APPWINDOW
CAPTION "AMF DVR Sample"
MENU IDC_DVR
FONT 8, "MS Shell Dlg", 400, 0, 0x1
BEGIN
    PUSHBUTTON      "Record",IDB_RECORD,62,20,50,27
    PUSHBUTTON      "Stop",IDB_STOP,142,20,50,27
    LTEXT           "Messages",IDC_STATIC,9,61,32,8
    EDITTEXT        IDE_MESSAGES,7,71,241,60,ES_MULTILINE | ES_AUTOHSCROLL | ES_READONLY | WS_VSCROLL
END


/////////////////////////////////////////////////////////////////////////////
//
// DESIGNINFO
//

#ifdef APSTUDIO_INVOKED
GUIDELINES DESIGNINFO
BEGIN
    IDD_MAIN_DIALOG, DIALOG
    BEGIN
        LEFTMARGIN, 7
        RIGHTMARGIN, 248
        TOPMARGIN, 7
        BOTTOMMARGIN, 143
    END
END
#endif    // APSTUDIO_INVOKED


/////////////////////////////////////////////////////////////////////////////
//
// AFX_DIALOG_LAYOUT
//

IDD_MAIN_DIALOG AFX_DIALOG_LAYOUT
BEGIN
    0
END

#endif    // English (Canada) resources
/////////////////////////////////////////////////////////////////////////////



#ifndef APSTUDIO_INVOKED
/////////////////////////////////////////////////////////////////////////////
//
// Generated from the TEXTINCLUDE 3 resource.
//


/////////////////////////////////////////////////////////////////////////////
#endif    // not APSTUDIO_INVOKED

