// Microsoft Visual C++ generated resource script.
//
#include "resource.h"

#define APSTUDIO_READONLY_SYMBOLS
/////////////////////////////////////////////////////////////////////////////
//
// Generated from the TEXTINCLUDE 2 resource.
//
#ifndef APSTUDIO_INVOKED
#include "targetver.h"
#endif
#define APSTUDIO_HIDDEN_SYMBOLS
#include "windows.h"
#undef APSTUDIO_HIDDEN_SYMBOLS

/////////////////////////////////////////////////////////////////////////////
#undef APSTUDIO_READONLY_SYMBOLS

/////////////////////////////////////////////////////////////////////////////
// English (United States) resources

#if !defined(AFX_RESOURCE_DLL) || defined(AFX_TARG_ENU)
LANGUAGE LANG_ENGLISH, SUBLANG_ENGLISH_US
#pragma code_page(1252)

/////////////////////////////////////////////////////////////////////////////
//
// Icon
//

// Icon with lowest ID value placed first to ensure application icon
// remains consistent on all systems.
IDI_PLAYBACKHW          ICON                    "PlaybackHW.ico"

IDI_SMALL               ICON                    "small.ico"


/////////////////////////////////////////////////////////////////////////////
//
// Menu
//

IDC_PLAYBACKHW MENU
BEGIN
    POPUP "&File"
    BEGIN
        MENUITEM "Open...\tCtrl+O",             ID_FILE_OPEN
        MENUITEM "Open Network Stream...\tCtrl+N", ID_FILE_OPENNETWORKSTREAM
        MENUITEM SEPARATOR
        MENUITEM "E&xit",                       IDM_EXIT
    END
    POPUP "&Options"
    BEGIN
        MENUITEM "VideoPresenter DX9",          ID_OPTIONS_PRESENTER_DX9
        MENUITEM "VideoPresenter DX11",         ID_OPTIONS_PRESENTER_DX11
        MENUITEM "VideoPresenter DX12",         ID_OPTIONS_PRESENTER_DX12
        MENUITEM "VideoPresenter OpenGL",       ID_OPTIONS_PRESENTER_OPENGL
        MENUITEM "VideoPresenter Vulkan",       ID_OPTIONS_PRESENTER_VULKAN
        MENUITEM SEPARATOR
        POPUP "HQ Scaler"
        BEGIN
            MENUITEM "Off",                         ID_HQSCALER_OFF
            MENUITEM SEPARATOR
            MENUITEM "Bi-linear\tCtrl+L",           ID_HQSCALER_BILINEAR
            MENUITEM "Bi-cubic\tCtrl+C",            ID_HQSCALER_BICUBIC
            MENUITEM "FSR\tCtrl+R",                 ID_HQSCALER_FSR
        END
        MENUITEM SEPARATOR
        MENUITEM "FRC",                         ID_OPTIONS_FRC
        MENUITEM SEPARATOR
        MENUITEM "Fullscreen\tCtrl+F",          ID_OPTIONS_FULLSCREEN
        MENUITEM "Exclusive Fullscreen",        ID_OPTIONS_EXCLUSIVEFULLSCREEN
    END
    MENUITEM "Play",                        ID_PLAYBACK_PLAY
    MENUITEM "Pause",                       ID_PLAYBACK_PAUSE
    MENUITEM "Step",                        ID_PLAYBACK_STEP
    MENUITEM "Stop",                        ID_PLAYBACK_STOP
    MENUITEM "Toolbar",                     ID_TOOLBAR
    POPUP "&Help"
    BEGIN
        MENUITEM "&About ...",                  IDM_ABOUT
    END
END


/////////////////////////////////////////////////////////////////////////////
//
// Accelerator
//

IDC_PLAYBACKHW ACCELERATORS
BEGIN
    "O",            ID_FILE_OPEN,           VIRTKEY, CONTROL, NOINVERT
    "N",            ID_FILE_OPENNETWORKSTREAM, VIRTKEY, CONTROL, NOINVERT
    "F",            ID_OPTIONS_FULLSCREEN,  VIRTKEY, CONTROL, NOINVERT
    "/",            IDM_ABOUT,              ASCII,  ALT, NOINVERT
    "?",            IDM_ABOUT,              ASCII,  ALT, NOINVERT
    "L",            ID_HQSCALER_BILINEAR,   VIRTKEY, CONTROL, NOINVERT
    "C",            ID_HQSCALER_BICUBIC,    VIRTKEY, CONTROL, NOINVERT
    "R",            ID_HQSCALER_FSR,        VIRTKEY, CONTROL, NOINVERT
END


/////////////////////////////////////////////////////////////////////////////
//
// Dialog
//

IDD_PROGRESS_DLG DIALOGEX 0, 0, 199, 86
STYLE DS_SETFONT | DS_MODALFRAME | DS_FIXEDSYS | WS_POPUP | WS_CAPTION | WS_SYSMENU
CAPTION "Playback"
FONT 8, "MS Shell Dlg", 0, 0, 0x1
BEGIN
    CONTROL         "",IDC_PROGRESS_BAR,"msctls_progress32",WS_BORDER,12,38,173,14
    LTEXT           "Optimization Components...",IDC_STATIC,12,21,90,8
    GROUPBOX        "Optimization Media SDK for your hardware",IDC_STATIC,4,4,189,57
    DEFPUSHBUTTON   "&Cancel",IDCANCEL,143,66,50,14
END

IDD_ABOUTBOX DIALOGEX 0, 0, 170, 62
STYLE DS_SETFONT | DS_MODALFRAME | DS_FIXEDSYS | WS_POPUP | WS_CAPTION | WS_SYSMENU
CAPTION "About PlaybackHW"
FONT 8, "MS Shell Dlg", 0, 0, 0x1
BEGIN
    ICON            IDR_MAINFRAME,IDC_STATIC,14,14,21,20
    LTEXT           "PlaybackHW, Version 1.0",IDC_STATIC,42,14,114,8,SS_NOPREFIX
    LTEXT           "Copyright (C) 2014",IDC_STATIC,42,26,114,8
    DEFPUSHBUTTON   "OK",IDOK,113,41,50,14,WS_GROUP
END

IDD_TOOLBAR_DLG DIALOGEX 0, 0, 310, 26
STYLE DS_SETFONT | DS_FIXEDSYS | WS_POPUP | WS_BORDER | WS_SYSMENU
FONT 8, "MS Shell Dlg", 400, 0, 0x1
BEGIN
    DEFPUSHBUTTON   "Close",IDOK,264,6,42,14
    CONTROL         "",IDC_SEEK,"msctls_trackbar32",TBS_BOTH | TBS_NOTICKS | WS_TABSTOP,6,6,246,15
END

IDD_OPEN_STREAM DIALOGEX 0, 0, 551, 69
STYLE DS_SETFONT | DS_MODALFRAME | DS_FIXEDSYS | WS_POPUP | WS_CAPTION | WS_SYSMENU
CAPTION "Dialog"
FONT 8, "MS Shell Dlg", 400, 0, 0x1
BEGIN
    EDITTEXT        IDC_EDIT_URL_VIDEO,66,6,480,14,ES_AUTOHSCROLL
    DEFPUSHBUTTON   "OK",IDOK,438,54,50,14
    PUSHBUTTON      "Cancel",IDCANCEL,498,54,50,14
    EDITTEXT        IDC_EDIT_URL_AUDIO,66,24,480,14,ES_AUTOHSCROLL
    LTEXT           "Video:",IDC_STATIC,12,12,21,8
    LTEXT           "Audio(opt):",IDC_STATIC,12,30,38,8
    CONTROL         "Listen as server",IDC_CHECK_LISTEN,"Button",BS_AUTOCHECKBOX | WS_TABSTOP,7,52,67,10
END


/////////////////////////////////////////////////////////////////////////////
//
// DESIGNINFO
//

#ifdef APSTUDIO_INVOKED
GUIDELINES DESIGNINFO
BEGIN
    IDD_PROGRESS_DLG, DIALOG
    BEGIN
        RIGHTMARGIN, 198
        BOTTOMMARGIN, 68
    END

    IDD_ABOUTBOX, DIALOG
    BEGIN
    END

    IDD_TOOLBAR_DLG, DIALOG
    BEGIN
        LEFTMARGIN, 7
        RIGHTMARGIN, 303
        TOPMARGIN, 7
        BOTTOMMARGIN, 19
    END

    IDD_OPEN_STREAM, DIALOG
    BEGIN
        LEFTMARGIN, 7
        RIGHTMARGIN, 544
        TOPMARGIN, 7
        BOTTOMMARGIN, 62
    END
END
#endif    // APSTUDIO_INVOKED


#ifdef APSTUDIO_INVOKED
/////////////////////////////////////////////////////////////////////////////
//
// TEXTINCLUDE
//

1 TEXTINCLUDE 
BEGIN
    "resource.h\0"
END

2 TEXTINCLUDE 
BEGIN
    "#ifndef APSTUDIO_INVOKED\r\n"
    "#include ""targetver.h""\r\n"
    "#endif\r\n"
    "#define APSTUDIO_HIDDEN_SYMBOLS\r\n"
    "#include ""windows.h""\r\n"
    "#undef APSTUDIO_HIDDEN_SYMBOLS\r\n"
    "\0"
END

3 TEXTINCLUDE 
BEGIN
    "\r\n"
    "\0"
END

#endif    // APSTUDIO_INVOKED


/////////////////////////////////////////////////////////////////////////////
//
// String Table
//

STRINGTABLE
BEGIN
    IDS_APP_TITLE           "PlaybackHW"
    IDC_PLAYBACKHW          "PLAYBACKHW"
END

#endif    // English (United States) resources
/////////////////////////////////////////////////////////////////////////////



#ifndef APSTUDIO_INVOKED
/////////////////////////////////////////////////////////////////////////////
//
// Generated from the TEXTINCLUDE 3 resource.
//


/////////////////////////////////////////////////////////////////////////////
#endif    // not APSTUDIO_INVOKED

