//***** The data in this file is reformatted from original work of MIT Media Laboratory *****//
//***** The original copyright notice is below:                                                 *****//

// This is a makehrtf HRIR definition file.  It is used to define the layout
// and source data to be processed into an OpenAL Soft compatible HRTF.
//
// This definition is used to transform the left ear HRIRs from the full set
// of KEMAR HRIRs provided by Bill Gardner <billg@media.mit.edu> and Keith
// Martin <kdm@media.mit.edu> of MIT Media Laboratory.
//
// The data full.tar.Z or full.zip is available from:
//
//  http://sound.media.mit.edu/resources/KEMAR.html
//
// It is copyright 1994 by MIT Media Laboratory, and provided free of charge
// with no restrictions on use so long as the authors above are cited.
//
// This definition is used to generate the internal HRTF table used by OpenAL
// Soft.

#pragma once

const amf_uint8 HRTF40e135aS44[] = 
{
	0x52, 0x49, 0x46, 0x46, 0x24, 0x02, 0x00, 0x00, 
	0x57, 0x41, 0x56, 0x45, 0x66, 0x6D, 0x74, 0x20, 
	0x10, 0x00, 0x00, 0x00, 0x01, 0x00, 0x02, 0x00, 
	0x44, 0xAC, 0x00, 0x00, 0x10, 0xB1, 0x02, 0x00, 
	0x04, 0x00, 0x10, 0x00, 0x64, 0x61, 0x74, 0x61, 
	0x00, 0x02, 0x00, 0x00, 0x01, 0x00, 0x06, 0x00, 
	0x01, 0x00, 0x06, 0x00, 0x01, 0x00, 0xF9, 0xFF, 
	0x01, 0x00, 0xF9, 0xFF, 0xFF, 0xFF, 0xF8, 0xFF, 
	0xFE, 0xFF, 0x03, 0x00, 0xFE, 0xFF, 0xFE, 0xFF, 
	0xFF, 0xFF, 0xFC, 0xFF, 0x02, 0x00, 0xFE, 0xFF, 
	0x00, 0x00, 0xDA, 0x00, 0xFE, 0xFF, 0x8D, 0x11, 
	0xFE, 0xFF, 0xD3, 0x1F, 0x00, 0x00, 0x87, 0xF2, 
	0x03, 0x00, 0x7E, 0xC7, 0xFF, 0xFF, 0x64, 0xE9, 
	0xFE, 0xFF, 0xDF, 0x0C, 0xFA, 0xFF, 0x2C, 0x2B, 
	0xFD, 0xFF, 0x9E, 0x41, 0xFB, 0xFF, 0x0D, 0x1D, 
	0x0A, 0x00, 0x32, 0xFE, 0xFB, 0xFF, 0x45, 0x0A, 
	0x03, 0x00, 0xB4, 0x16, 0x40, 0x00, 0xEC, 0x01, 
	0xD3, 0x04, 0x67, 0xE9, 0x8F, 0x0B, 0xC8, 0xEB, 
	0x70, 0x03, 0xA4, 0xEC, 0x90, 0xEE, 0x38, 0xE9, 
	0xD4, 0xE9, 0x14, 0xED, 0xB7, 0xFC, 0x28, 0xF3, 
	0x7F, 0x10, 0x7B, 0xF0, 0xB7, 0x18, 0x25, 0xF5, 
	0x4A, 0x14, 0xAD, 0xFE, 0xFF, 0x09, 0x03, 0x04, 
	0xC7, 0x06, 0xC1, 0x04, 0x9C, 0x08, 0x0C, 0x06, 
	0x06, 0x06, 0x04, 0x0D, 0x13, 0x00, 0x59, 0x0A, 
	0xD7, 0xFD, 0xD2, 0x05, 0xEB, 0xFB, 0xA5, 0x05, 
	0x69, 0xF8, 0x93, 0x02, 0xDE, 0xF6, 0x58, 0xFE, 
	0x13, 0xF7, 0xB0, 0x01, 0xA6, 0xF7, 0x37, 0xFF, 
	0x80, 0xF8, 0x7F, 0xFA, 0x4D, 0xFB, 0x38, 0xFA, 
	0xB7, 0xFC, 0x78, 0xFB, 0x0D, 0xFE, 0x44, 0xFC, 
	0x41, 0x03, 0xE4, 0xFA, 0xAB, 0x05, 0x7C, 0xFA, 
	0xA9, 0x05, 0x23, 0xFA, 0xBE, 0x05, 0x01, 0xFD, 
	0x34, 0x04, 0x47, 0xFF, 0x30, 0x04, 0x23, 0xFF, 
	0x89, 0x03, 0xB1, 0xFF, 0x66, 0x01, 0xB8, 0x01, 
	0x8E, 0xFE, 0x34, 0x03, 0xC0, 0xFB, 0x9E, 0x02, 
	0x90, 0xFB, 0xCA, 0x00, 0x21, 0xFD, 0xA0, 0xFE, 
	0x72, 0xFD, 0xA7, 0xFD, 0x08, 0xFC, 0xAA, 0xFD, 
	0x89, 0xFB, 0xCD, 0xFD, 0x5D, 0xFD, 0x03, 0xFE, 
	0x37, 0xFF, 0x02, 0xFE, 0x21, 0xFF, 0xF6, 0xFE, 
	0x36, 0xFF, 0x55, 0xFF, 0xE9, 0xFF, 0x6F, 0xFF, 
	0xBC, 0x00, 0x70, 0xFF, 0x27, 0x01, 0xBD, 0xFF, 
	0xB6, 0x00, 0x1F, 0x00, 0x8B, 0x00, 0x43, 0x00, 
	0xBF, 0x00, 0xE3, 0xFF, 0x75, 0x00, 0x43, 0xFF, 
	0xD7, 0xFF, 0xB5, 0xFF, 0x31, 0xFF, 0x18, 0x00, 
	0x66, 0xFE, 0xD2, 0xFF, 0xDE, 0xFD, 0xFF, 0xFE, 
	0xAF, 0xFD, 0xE9, 0xFE, 0x88, 0xFD, 0x1D, 0xFF, 
	0x57, 0xFD, 0x61, 0xFF, 0x7D, 0xFD, 0xA7, 0xFF, 
	0x5F, 0xFE, 0xE1, 0xFF, 0xF4, 0xFE, 0xBB, 0xFF, 
	0x4F, 0xFF, 0x88, 0xFF, 0xD1, 0xFF, 0x6C, 0xFF, 
	0x04, 0x00, 0x4D, 0xFF, 0x20, 0x00, 0x44, 0xFF, 
	0x05, 0x00, 0xA8, 0xFF, 0xFF, 0xFF, 0x15, 0x00, 
	0xCF, 0xFF, 0x10, 0x00, 0x74, 0xFF, 0xD5, 0xFF, 
	0x1E, 0xFF, 0x99, 0xFF, 0xBD, 0xFE, 0x74, 0xFF, 
	0xBD, 0xFE, 0x92, 0xFF, 0x08, 0xFF, 0xEF, 0xFF, 
	0x3E, 0xFF, 0xEE, 0xFF, 0x21, 0xFF, 0x88, 0xFF, 
	0x0C, 0xFF, 0x0E, 0xFF, 0x3A, 0xFF, 0x05, 0xFF, 
	0x5C, 0xFF, 0x2F, 0xFF, 0x67, 0xFF, 0x51, 0xFF, 
	0x7A, 0xFF, 0x2C, 0xFF, 0xA2, 0xFF, 0x37, 0xFF, 
	0xC3, 0xFF, 0x6B, 0xFF, 0xD3, 0xFF, 0xA3, 0xFF, 
	0xE0, 0xFF, 0xD4, 0xFF, 0xE0, 0xFF, 0xD1, 0xFF, 
	0xE6, 0xFF, 0xB2, 0xFF, 0xD6, 0xFF, 0xA9, 0xFF, 
	0xAB, 0xFF, 0xBD, 0xFF, 0x8D, 0xFF, 0xB7, 0xFF, 
	0x92, 0xFF, 0xAE, 0xFF, 0x9A, 0xFF, 0xB3, 0xFF, 
	0x7C, 0xFF, 0xBA, 0xFF, 0x6A, 0xFF, 0x93, 0xFF, 
	0x82, 0xFF, 0x73, 0xFF, 0xB0, 0xFF, 0x71, 0xFF, 
	0xDC, 0xFF, 0x6E, 0xFF, 0xD2, 0xFF, 0x5E, 0xFF, 
	0x9E, 0xFF, 0x73, 0xFF, 0x84, 0xFF, 0xA0, 0xFF, 
	0xAF, 0xFF, 0xBA, 0xFF, 0xE1, 0xFF, 0xD7, 0xFF, 
	0xEA, 0xFF, 0x20, 0x00, 0xCD, 0xFF, 0x57, 0x00, 
	0x9E, 0xFF, 0x61, 0x00, 0x92, 0xFF, 0x59, 0x00, 
	0xA3, 0xFF, 0x3E, 0x00
};

const amf_size HRTF40e135aS44Count=556;

