//***** The data in this file is reformatted from original work of MIT Media Laboratory *****//
//***** The original copyright notice is below:                                                 *****//

// This is a makehrtf HRIR definition file.  It is used to define the layout
// and source data to be processed into an OpenAL Soft compatible HRTF.
//
// This definition is used to transform the left ear HRIRs from the full set
// of KEMAR HRIRs provided by Bill Gardner <billg@media.mit.edu> and Keith
// Martin <kdm@media.mit.edu> of MIT Media Laboratory.
//
// The data full.tar.Z or full.zip is available from:
//
//  http://sound.media.mit.edu/resources/KEMAR.html
//
// It is copyright 1994 by MIT Media Laboratory, and provided free of charge
// with no restrictions on use so long as the authors above are cited.
//
// This definition is used to generate the internal HRTF table used by OpenAL
// Soft.

#pragma once

const amf_uint8 HRTF0e160aS44[] = 
{
	0x52, 0x49, 0x46, 0x46, 0x24, 0x02, 0x00, 0x00, 
	0x57, 0x41, 0x56, 0x45, 0x66, 0x6D, 0x74, 0x20, 
	0x10, 0x00, 0x00, 0x00, 0x01, 0x00, 0x02, 0x00, 
	0x44, 0xAC, 0x00, 0x00, 0x10, 0xB1, 0x02, 0x00, 
	0x04, 0x00, 0x10, 0x00, 0x64, 0x61, 0x74, 0x61, 
	0x00, 0x02, 0x00, 0x00, 0xFD, 0xFF, 0x0E, 0x00, 
	0xFE, 0xFF, 0xED, 0xFF, 0xFF, 0xFF, 0x1E, 0x00, 
	0xFE, 0xFF, 0xDC, 0xFF, 0xFD, 0xFF, 0x20, 0x00, 
	0xFC, 0xFF, 0xD7, 0xFF, 0x02, 0x00, 0x22, 0x00, 
	0x00, 0x00, 0xD8, 0xFF, 0x06, 0x00, 0x27, 0x00, 
	0xF9, 0xFF, 0xC8, 0xFF, 0x0D, 0x00, 0x15, 0x08, 
	0xEE, 0xFF, 0x7A, 0x1B, 0x12, 0x00, 0x99, 0xFC, 
	0xE6, 0xFF, 0x06, 0xEC, 0x21, 0x00, 0x7E, 0x06, 
	0xE1, 0xFF, 0x4E, 0x00, 0x3E, 0x00, 0xE6, 0xF5, 
	0x1F, 0x00, 0xE1, 0x11, 0x1C, 0x08, 0x21, 0x2F, 
	0xBA, 0x11, 0x85, 0x17, 0x4A, 0xFD, 0x33, 0x0B, 
	0xE9, 0xF0, 0x19, 0x17, 0x7D, 0x00, 0xEC, 0x03, 
	0x3E, 0x04, 0xC2, 0xF2, 0xAC, 0xFB, 0xF9, 0xF5, 
	0x89, 0x0D, 0xB0, 0xEC, 0x7B, 0x26, 0x85, 0xE5, 
	0xB5, 0x11, 0xC1, 0xEC, 0x46, 0xFE, 0xFF, 0xED, 
	0x52, 0x09, 0x4A, 0xEF, 0xAF, 0x05, 0x9C, 0xF6, 
	0xC3, 0xF6, 0x44, 0xFD, 0x87, 0xF6, 0x9E, 0x02, 
	0x6C, 0xF5, 0x98, 0x04, 0xAA, 0xED, 0xAF, 0x0A, 
	0xFE, 0xF1, 0x8A, 0x0B, 0xEA, 0xF9, 0xD7, 0x05, 
	0xED, 0xF7, 0x12, 0x04, 0x04, 0xF8, 0xC9, 0x02, 
	0x15, 0x00, 0x78, 0xFF, 0xE0, 0x03, 0x84, 0xFC, 
	0x95, 0x01, 0xB5, 0xFC, 0xE8, 0x04, 0xB9, 0xFC, 
	0x55, 0x08, 0xC3, 0xFC, 0x2E, 0x05, 0xAB, 0xFB, 
	0x85, 0x01, 0xB6, 0xFA, 0x74, 0x01, 0x6A, 0xF8, 
	0x38, 0x01, 0x1E, 0xF9, 0x03, 0xFF, 0x3B, 0xFC, 
	0xD6, 0xFE, 0x9C, 0xFD, 0x35, 0xFF, 0x66, 0xFF, 
	0xB3, 0xFD, 0x3E, 0x01, 0x8C, 0xFC, 0x5A, 0x02, 
	0xAB, 0xFC, 0xEA, 0x02, 0xB6, 0xFB, 0x95, 0x03, 
	0x4C, 0xFC, 0x0B, 0x04, 0x75, 0xFF, 0x28, 0x03, 
	0x39, 0x01, 0xEF, 0x00, 0xE1, 0x00, 0xEB, 0xFF, 
	0x5D, 0x00, 0x9B, 0xFF, 0xDE, 0x00, 0x6A, 0xFE, 
	0xC4, 0x00, 0xA1, 0xFD, 0xB9, 0xFF, 0x33, 0xFE, 
	0x26, 0xFF, 0x5F, 0xFE, 0x9C, 0xFE, 0x31, 0xFE, 
	0x15, 0xFE, 0x51, 0xFE, 0xFB, 0xFD, 0xF1, 0xFE, 
	0xF2, 0xFD, 0x3C, 0xFF, 0x27, 0xFE, 0x65, 0xFF, 
	0xE3, 0xFE, 0x53, 0xFF, 0xAD, 0xFF, 0x1D, 0xFF, 
	0xBC, 0xFF, 0x29, 0xFF, 0x2A, 0xFF, 0x62, 0xFF, 
	0x5F, 0xFF, 0xEC, 0xFF, 0x67, 0x00, 0x93, 0x00, 
	0x9B, 0x00, 0x6E, 0x00, 0x15, 0x00, 0x02, 0x00, 
	0xDA, 0xFF, 0x08, 0x00, 0xE3, 0xFF, 0x5C, 0xFF, 
	0xAD, 0xFF, 0x97, 0xFE, 0x6A, 0xFF, 0xDC, 0xFE, 
	0x89, 0xFF, 0x61, 0xFF, 0xEB, 0xFF, 0x3F, 0xFF, 
	0xB9, 0xFF, 0x67, 0xFF, 0x31, 0xFF, 0x6F, 0xFF, 
	0xE8, 0xFE, 0x11, 0xFF, 0x8F, 0xFE, 0xEC, 0xFE, 
	0x8E, 0xFE, 0xAD, 0xFF, 0x2E, 0xFF, 0x45, 0x00, 
	0xA6, 0xFF, 0x42, 0x00, 0x71, 0xFF, 0x5A, 0x00, 
	0x43, 0xFF, 0x44, 0x00, 0x94, 0xFF, 0x24, 0x00, 
	0xC1, 0xFF, 0x34, 0x00, 0xB0, 0xFF, 0x38, 0x00, 
	0xF5, 0xFF, 0x00, 0x00, 0x36, 0x00, 0xB8, 0xFF, 
	0x25, 0x00, 0x81, 0xFF, 0xD4, 0xFF, 0x52, 0xFF, 
	0x89, 0xFF, 0x1F, 0xFF, 0x56, 0xFF, 0x12, 0xFF, 
	0x6E, 0xFF, 0x00, 0xFF, 0x9F, 0xFF, 0xE0, 0xFE, 
	0x88, 0xFF, 0x97, 0xFE, 0x42, 0xFF, 0x98, 0xFE, 
	0x04, 0xFF, 0xEF, 0xFE, 0x22, 0xFF, 0x60, 0xFF, 
	0x72, 0xFF, 0xBE, 0xFF, 0xBF, 0xFF, 0xEB, 0xFF, 
	0xCB, 0xFF, 0xD7, 0xFF, 0x80, 0xFF, 0x94, 0xFF, 
	0x22, 0xFF, 0x8A, 0xFF, 0x18, 0xFF, 0xA5, 0xFF, 
	0x7D, 0xFF, 0xBA, 0xFF, 0xE1, 0xFF, 0xBE, 0xFF, 
	0xF6, 0xFF, 0x9C, 0xFF, 0xD2, 0xFF, 0x6F, 0xFF, 
	0x94, 0xFF, 0x76, 0xFF, 0x71, 0xFF, 0xA2, 0xFF, 
	0x81, 0xFF, 0xDE, 0xFF, 0xA4, 0xFF, 0xFF, 0xFF, 
	0xA9, 0xFF, 0xFB, 0xFF, 0x89, 0xFF, 0xD4, 0xFF, 
	0x6C, 0xFF, 0xCD, 0xFF, 0x66, 0xFF, 0xD3, 0xFF, 
	0x87, 0xFF, 0xD3, 0xFF, 0xCA, 0xFF, 0xED, 0xFF, 
	0x0D, 0x00, 0xE9, 0xFF
};

const amf_size HRTF0e160aS44Count=556;

